package com.mcsnowflake.worktimer

import android.content.ComponentName
import android.content.Context
import android.content.ServiceConnection
import android.media.AudioManager
import android.os.Bundle
import android.os.IBinder
import android.util.Log
import androidx.activity.ComponentActivity
import androidx.activity.compose.setContent
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.ui.Modifier
import androidx.datastore.core.DataStore
import androidx.datastore.preferences.core.Preferences
import androidx.datastore.preferences.core.booleanPreferencesKey
import androidx.datastore.preferences.core.edit
import androidx.datastore.preferences.core.stringPreferencesKey
import androidx.lifecycle.coroutineScope
import androidx.lifecycle.lifecycleScope
import androidx.navigation.NavType
import androidx.navigation.compose.NavHost
import androidx.navigation.compose.composable
import androidx.navigation.compose.rememberNavController
import androidx.navigation.navArgument
import com.mcsnowflake.worktimer.service.TimerService
import com.mcsnowflake.worktimer.service.TimerState
import com.mcsnowflake.worktimer.ui.DASHBOARD
import com.mcsnowflake.worktimer.ui.DashBoardViewModel
import com.mcsnowflake.worktimer.ui.Dashboard
import com.mcsnowflake.worktimer.ui.about.ABOUT_SCREEN
import com.mcsnowflake.worktimer.ui.about.AboutScreen
import com.mcsnowflake.worktimer.ui.about.showAppBar
import com.mcsnowflake.worktimer.ui.settings.Settings
import com.mcsnowflake.worktimer.ui.theme.Theme
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.Job
import kotlinx.coroutines.flow.collect
import kotlinx.coroutines.flow.first
import kotlinx.coroutines.launch
import kotlinx.coroutines.runBlocking
import org.koin.android.ext.android.inject
import org.koin.androidx.viewmodel.ext.android.viewModel

const val PREFERENCES_SCREEN = "Preferences"
val FIRST_START = booleanPreferencesKey("FIRST_START")
val VERSION = stringPreferencesKey("VERSION")

class MainActivity : ComponentActivity() {

    private val timerViewModel by viewModel<DashBoardViewModel>()
    private val Context.dataStore: DataStore<Preferences> by inject()

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)

        checkForOldSettingsAndConvertToDataStore()

        bindService() // sending the binding request here lets it bind as soon as the service is up

        val startScreen = selectStartScreen()

        setContent {
            Theme {
                val navController = rememberNavController()
                Column(Modifier.fillMaxSize()) {
                    NavHost(navController, startScreen) {
                        composable(ABOUT_SCREEN, listOf(navArgument(showAppBar) { type = NavType.BoolType })) {
                            AboutScreen(
                                navController,
                                it.arguments?.getBoolean(showAppBar) ?: false
                            )
                        }
                        composable(DASHBOARD) {
                            Dashboard(
                                navController,
                                dashBoardViewModel = timerViewModel,
                                audioManager = getSystemService(AUDIO_SERVICE) as AudioManager
                            )
                        }
                        composable(PREFERENCES_SCREEN) { Settings(this@MainActivity, navController) }
                    }
                }
            }
        }
    }

    private fun selectStartScreen(): String {
        val firstStart = runBlocking { !dataStore.data.first().contains(FIRST_START) }
        val storedVersion = runBlocking { dataStore.data.first()[VERSION] ?: "" }
        val currentVersion = BuildConfig.VERSION_NAME.substringBefore('-')
        if (storedVersion != currentVersion) lifecycleScope.launch { dataStore.edit { it[VERSION] = currentVersion } }

        return if (firstStart) {
            lifecycleScope.launch { dataStore.edit { it[FIRST_START] = false } }
            ABOUT_SCREEN
        } else DASHBOARD
    }

    private val connection: ServiceConnection = object : ServiceConnection {
        private lateinit var eventsConnector: Job
        override fun onServiceConnected(name: ComponentName?, service: IBinder?) {
            val binder = service as TimerService.LocalBinder
            eventsConnector = lifecycle.coroutineScope.launch {
                binder.getTimerEvents().collect {

                    Log.d("MainActivity", it.toString())
                    timerViewModel.timerState.value = it
                }
            }
            Log.d("MainActivity", "timer service connected")
        }

        override fun onServiceDisconnected(name: ComponentName?) {
            eventsConnector.cancel()
            lifecycle.coroutineScope.launch { timerViewModel.timerState.value = TimerState.NotRunning }
            Log.d("MainActivity", "timer service disconnected")
        }

        override fun onBindingDied(name: ComponentName?) {
            super.onBindingDied(name)
            bindService()
        }
    }

    private fun bindService() {
        bindService(TimerService.getIntent(applicationContext), connection, 0)
    }

    override fun onDestroy() {
        super.onDestroy()
        unbindService(connection)
    }

    override fun onResume() {
        super.onResume()
        // TODO preferencesVM.recheckDND()
    }

    private fun checkForOldSettingsAndConvertToDataStore() = lifecycle.coroutineScope.launch(Dispatchers.IO) {
        applicationContext.getSharedPreferences(applicationContext.packageName + "_preferences", MODE_PRIVATE).recoverTo(applicationContext.dataStore)
    }
}
