package com.mcsnowflake.worktimer.ui.settings

import android.app.Activity
import android.app.NotificationManager
import android.content.Intent
import android.provider.Settings
import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.Spacer
import androidx.compose.foundation.layout.fillMaxHeight
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.height
import androidx.compose.foundation.layout.heightIn
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.width
import androidx.compose.material.AlertDialog
import androidx.compose.material.Button
import androidx.compose.material.ContentAlpha
import androidx.compose.material.Icon
import androidx.compose.material.LocalContentAlpha
import androidx.compose.material.MaterialTheme
import androidx.compose.material.ProvideTextStyle
import androidx.compose.material.Surface
import androidx.compose.material.Switch
import androidx.compose.material.Text
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.rounded.DoNotDisturb
import androidx.compose.material.icons.rounded.MoreTime
import androidx.compose.material.icons.rounded.Redo
import androidx.compose.material.icons.rounded.WaterDrop
import androidx.compose.runtime.Composable
import androidx.compose.runtime.CompositionLocalProvider
import androidx.compose.runtime.collectAsState
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.vector.ImageVector
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import androidx.navigation.NavHostController
import com.mcsnowflake.worktimer.AUTO_SWITCH_MODE
import com.mcsnowflake.worktimer.DND_MODE
import com.mcsnowflake.worktimer.HYDRATION_MODE
import com.mcsnowflake.worktimer.OVERTIME_MODE
import com.mcsnowflake.worktimer.ui.shared.AppBar
import com.mcsnowflake.worktimer.ui.shared.AppBarStatus
import org.koin.androidx.compose.get
import org.koin.androidx.compose.getViewModel

@Composable
fun Settings(
    activity: Activity,
    navController: NavHostController,
    preferences: SettingsViewModel = getViewModel()
) = Surface(
    modifier = Modifier.fillMaxSize(),
    color = MaterialTheme.colors.background
) {
    Column(modifier = Modifier.fillMaxSize()) {

        AppBar(navController, AppBarStatus(title = "Settings", showUpArrow = true))

        val notificationManager: NotificationManager = get()
        val preferencesFullyLoaded by preferences.allLoaded.collectAsState()

        var showDndDialog by remember { mutableStateOf(false) }
        if (showDndDialog) DndPermissionDialog(activity) { showDndDialog = false }

        Column(modifier = Modifier.padding(16.dp)) {

            val autoSwitchMode by preferences[AUTO_SWITCH_MODE]
            SettingsCard(
                icon = Icons.Rounded.Redo,
                label = "Automatic session switching",
                secText = if (autoSwitchMode) "App switches automatically to the next session"
                else "App waits for your approval to switch to the next session",
                state = autoSwitchMode,
                onChange = { preferences[AUTO_SWITCH_MODE] = it },
                enabled = preferencesFullyLoaded
            )
            val overtimeOption by preferences[OVERTIME_MODE]
            SettingsCard(
                icon = Icons.Rounded.MoreTime,
                label = "Optional session prolonging",
                secText = if (overtimeOption) "An additional button allows to extend the current session by 5 min"
                else "Session time is fix once the timer has started",
                state = overtimeOption,
                onChange = { preferences[OVERTIME_MODE] = it },
                enabled = preferencesFullyLoaded
            )

            val dndMode by preferences[DND_MODE]
            SettingsCard(
                icon = Icons.Rounded.DoNotDisturb,
                label = "No Distractions",
                secText = if (dndMode) "Do Not Disturb mode will be enabled during work sessions"
                else "Do Not Disturb will not be enabled during work sessions",
                state = dndMode,
                onChange = {
                    when {
                        it && !notificationManager.isNotificationPolicyAccessGranted -> showDndDialog = true
                        else -> preferences[DND_MODE] = it
                    }
                }, // preferences::changeDndMode
                enabled = preferencesFullyLoaded
            )
            val hydrationMode by preferences[HYDRATION_MODE]
            SettingsCard(
                icon = Icons.Rounded.WaterDrop,
                label = "Hydration reminders",
                secText = if (hydrationMode) "Regular reminders to drink water will be issued"
                else "Hydration reminders are disabled",
                state = hydrationMode,
                onChange = { preferences[HYDRATION_MODE] = it },
                enabled = preferencesFullyLoaded
            )
        }
    }
}

@Composable
fun DndPermissionDialog(activity: Activity, onDismiss: () -> Unit) {
    AlertDialog(
        onDismissRequest = { onDismiss() },
        dismissButton = { Button({ onDismiss() }) { Text("Cancel") } },
        confirmButton = {
            Button({
                onDismiss()
                activity.startActivity(Intent(Settings.ACTION_NOTIFICATION_POLICY_ACCESS_SETTINGS))
            }) { Text("Ok") }
        },
        title = { Text("Permissions required") },
        text = { Text("In order to change the DND mode, the app needs according permissions. You will be forwarded to the settings page.") }
    )
}

@Composable
fun SettingsCard(
    modifier: Modifier = Modifier,
    icon: ImageVector? = null,
    label: String,
    secText: String,
    state: Boolean,
    onChange: (state: Boolean) -> Unit,
    enabled: Boolean
) = Row(
    modifier = modifier
        .fillMaxWidth()
        .height(SettingsCardHeight),
    verticalAlignment = Alignment.CenterVertically,
    horizontalArrangement = Arrangement.spacedBy(16.dp)
) {

    if (icon == null) {
        Spacer(Modifier.width(SettingsCardInset))
    } else {
        Row(
            Modifier
                .fillMaxHeight()
                .width(SettingsCardInset),
            Arrangement.Center,
            Alignment.CenterVertically
        ) {
            CompositionLocalProvider(LocalContentAlpha provides ContentAlpha.high) { Icon(icon, "$label icon") }
        }
    }

    Column(
        Modifier
            .fillMaxHeight()
            .weight(1f),
        verticalArrangement = Arrangement.Center
    ) {
        ProvideTextStyle(value = MaterialTheme.typography.body1) {
            CompositionLocalProvider(
                LocalContentAlpha provides ContentAlpha.high
            ) { Text(text = label) }
        }
        ProvideTextStyle(value = MaterialTheme.typography.body2) {
            CompositionLocalProvider(
                LocalContentAlpha provides ContentAlpha.high
            ) { Text(text = secText, color = Color.Gray, fontSize = 12.sp, modifier = Modifier.heightIn(min = 25.dp, max = 40.dp)) }
        }
    }
    CompositionLocalProvider(LocalContentAlpha provides ContentAlpha.medium) {
        Row(
            Modifier.fillMaxHeight(),
            horizontalArrangement = Arrangement.End,
            verticalAlignment = Alignment.CenterVertically
        ) {
            Switch(
                checked = state,
                onCheckedChange = onChange,
                enabled = enabled
            )
        }
    }
}

private val SettingsCardHeight = 80.dp
private val SettingsCardInset = 40.dp
