package com.mcsnowflake.worktimer.ui.shared

import androidx.compose.foundation.isSystemInDarkTheme
import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.wrapContentSize
import androidx.compose.material.AlertDialog
import androidx.compose.material.Button
import androidx.compose.material.Divider
import androidx.compose.material.DropdownMenu
import androidx.compose.material.DropdownMenuItem
import androidx.compose.material.Icon
import androidx.compose.material.IconButton
import androidx.compose.material.MaterialTheme
import androidx.compose.material.SnackbarDefaults
import androidx.compose.material.Text
import androidx.compose.material.TopAppBar
import androidx.compose.material.contentColorFor
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.MoreVert
import androidx.compose.material.icons.rounded.ArrowBack
import androidx.compose.runtime.Composable
import androidx.compose.runtime.MutableState
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.vector.ImageVector
import androidx.compose.ui.platform.LocalUriHandler
import androidx.compose.ui.unit.dp
import androidx.navigation.NavHostController
import com.mcsnowflake.worktimer.ui.about.ABOUT_SCREEN
import com.mcsnowflake.worktimer.ui.about.showAppBar

@Composable
fun AppBar(navController: NavHostController, appStatus: AppBarStatus) {
    TopAppBar(
        title = { Text(appStatus.title) },
        actions = {
            appStatus.actionIcons(navController)
            if (appStatus.showMenu) DropdownMenu(navController)
        },
        navigationIcon = {
            if (appStatus.showUpArrow) {
                IconButton({ navController.popBackStack() }) { Icon(Icons.Rounded.ArrowBack, "Back") }
            } else {
                appStatus.navIcon?.let { Icon(it, null, Modifier.padding(start = 16.dp)) }
            }
        },
        contentColor = if (isSystemInDarkTheme()) MaterialTheme.colors.primary else contentColorFor(SnackbarDefaults.backgroundColor)
    )
}

data class AppBarStatus(
    val title: String,
    val actionIcons: @Composable (NavHostController) -> Unit = {},
    val navIcon: ImageVector? = null,
    val showMenu: Boolean = false,
    val showUpArrow: Boolean = false
)

@Composable
fun DropdownMenu(navController: NavHostController) {
    val expandMenu = remember { mutableStateOf(false) }
    val showBugReportDialog = remember { mutableStateOf(false) }

    Box(
        Modifier.wrapContentSize(Alignment.TopEnd)
    ) {
        IconButton({ expandMenu.value = true }) { Icon(Icons.Filled.MoreVert, "More Menu") }
    }

    DropdownMenu(
        expanded = expandMenu.value,
        onDismissRequest = { expandMenu.value = false },
    ) {
        DropdownMenuItem(onClick = {
            expandMenu.value = false
            navController.navigate(ABOUT_SCREEN.showAppBar())
        }) {
            Text("About the app")
        }
        Divider()
        DropdownMenuItem(onClick = {
            expandMenu.value = false
            showBugReportDialog.value = true
        }) {
            Text("Report a bug")
        }
    }

    BugReportDialog(showBugReportDialog)
}

@Composable
fun BugReportDialog(showDialog: MutableState<Boolean>) {
    if (showDialog.value) {
        val uriHandler = LocalUriHandler.current
        AlertDialog(
            onDismissRequest = {
            },
            title = {
                Text("Report a Bug")
            },
            confirmButton = {
                Button(
                    onClick = {
                        showDialog.value = false
                        uriHandler.openUri("https://gitlab.com/McSnowflake/worktimer/-/issues")
                    },
                ) {
                    Text("Ok")
                }
            },
            dismissButton = {
                Button(
                    onClick = { showDialog.value = false },
                ) {
                    Text("Decline")
                }
            },
            text = {
                Column(
                    verticalArrangement = Arrangement.spacedBy(8.dp)
                ) {
                    Text("You will be forwarded to GitLab, where bugs are tracked for this project.")
                    Text("In order to file a bug, you need to be registered.")
                    Text("Please check if there already is a bug reporting your problem. If so, please add to it instead of creating a new one.")
                }
            },
        )
    }
}
